package soal.ppr

import java.util.Random

import com.twitter.cassovary.graph.{DirectedGraph, Node}
import soal.ppr.experiments.ExperimentUtils
import soal.util._

import scala.collection.mutable

/**
 * Contains methods related to personalized PageRank estimation on undirected graphs.  All methods operate in the
 * context of the given graph, teleportProbability, and random.
 */

class BidirectionalUndirectedPPREstimator (val graph: DirectedGraph[Node],
                                 val teleportProbability: Float,
                                 val random: Random = new Random) {
  /** The number of milliseconds it takes to generate a random walk.  Could be tuned on a given platform and graph to
    * improve running time. */
  val msPerWalk = 5.0e-4f

  /**
   * Estimates the personalized PageRank from the given source to the given target on the given undirected
   * graph.  The result will have mean relative error less than the given bound.
   */
  def estimatePPR(startId: Int,
                  targetId: Int,
                  minimumPPR: Float,
                  relativeError: Float,
                  edgeWeight: (Node, Node) => Float = (u: Node, v: Node) => 1.0f / u.outboundCount,
                  guaranteeRelativeError:Boolean = false): Float = {
    val target = graph.getNodeById(targetId).get
    val chernoffConstant = if (guaranteeRelativeError)
      3 * math.log(2 / 1.0e-9) // guarantees the given relative error with probability 1 - 1.0e-9
    else
      0.07 // Found in experiments to give roughly correct mean relative error when
    // relativeError in {5%, 10%, 20%}

    val chernoffMultiplier = chernoffConstant / math.pow(relativeError, 2.0).toFloat

    def computeWalkCount(maxResidualRatio: Float): Int =
      (chernoffMultiplier * target.outboundCount * maxResidualRatio / minimumPPR).toInt
    def estimateForwardTimeInMillis(maxResidual: Float): Float =
      computeWalkCount(maxResidual) * msPerWalk

    val ((estimates, residualRatios, maxResidualRatio), reverseTime) = ExperimentUtils.time {
      balancedForwardResiduals(startId, estimateForwardTimeInMillis)
    }

    val walkCount = computeWalkCount(maxResidualRatio)
    var estimate = estimates(targetId)
    val (_, forwardTime) = ExperimentUtils.time {
     for (walkIndex <- 0 until walkCount) {
        val v = samplePPR(targetId)
        estimate += target.outboundCount * residualRatios(v) / walkCount
      }
    }
    println(s"walkCount maxResidualRatio forwardTime reverseTime $walkCount $maxResidualRatio $forwardTime $reverseTime")
    estimate
  }

  /** Returns (estimates, residualRatios, maxResidualDegreeRatio) */
  def balancedForwardResiduals(startId: Int,
                               walkMillisGivenMaxResidualRatio: Float => Double):
      (collection.Map[Int, Float], collection.Map[Int, Float], Float) = {
    val pprEstimates = CollectionsUtil.efficientIntFloatMapWithDefault0() // pprEstimates(vId) estimates ppr(s, vId)
    val residualRatioQueue = new HeapMappedIntPriorityQueue() // Gives each node v priority residual(v) / outDegree(v)

    val startOutDegree = graph.getNodeById(startId).get.outboundCount
    residualRatioQueue.insert(startId, 1.0f / startOutDegree)

    val startTime = System.currentTimeMillis()
    def elapsedTime(): Double = System.currentTimeMillis() - startTime

    while (!residualRatioQueue.isEmpty &&
      elapsedTime() < walkMillisGivenMaxResidualRatio(residualRatioQueue.maxPriority)) {
      val uResidualRatio = residualRatioQueue.maxPriority
      val uId = residualRatioQueue.extractMax()
      val u = graph.getNodeById(uId).get
      val uResidual = uResidualRatio * u.outboundCount
      pprEstimates(uId) = pprEstimates(uId) + uResidual * teleportProbability

      for (vId <- u.outboundNodes()) {
        val v = graph.getNodeById(vId).get
        // Don't need to divide by u.outboundCount because uResidualRatio already incorporates that
        val deltaResidual = (1.0f - teleportProbability)  * uResidualRatio
        val deltaResidualRatio = deltaResidual / v.outboundCount
        if (! residualRatioQueue.contains(vId)) {
          residualRatioQueue.insert(vId, 0.0f)
        }
        residualRatioQueue.increasePriority(vId, residualRatioQueue.getPriority(vId) + deltaResidualRatio)
      }
    }

    val residualRatios = residualRatioQueue.currentPriorities()
    val maxResidualDegreeRatio = if (residualRatioQueue.isEmpty) 0.0f else residualRatioQueue.maxPriority
    (pprEstimates, residualRatios, maxResidualDegreeRatio)
  }

  /**
   * Samples from the PageRank distribution personalized to the given source by doing a single
   * random walk.
   */
  def samplePPR(sourceId: Int): Int = {
    val sourceDistribution = new ConstantDistribution(sourceId)
    samplePPR(sourceDistribution)
  }

  /**
   * Samples from the PageRank distribution personalized to the given source distribution by
   * doing a single random walk.
   */
  def samplePPR(sourceDistribution: DiscreteDistribution): Int = {
    var v = sourceDistribution.sample()
    while(random.nextFloat() > teleportProbability) {
      val vNode = graph.getNodeById(v).get
      vNode.randomOutboundNode(random) match {
        case Some(outNeighbor) =>
          v = outNeighbor
        case None =>
          v = sourceDistribution.sample() // Teleport after reaching dead end.
      }
    }
    v
  }
}
