package soal.ppr.experiments

import com.twitter.cassovary.graph.{Node, DirectedGraph}
import java.util.Random
import soal.ppr.{experiments, BidirectionalPPRPrecomputor, BidirectionalPPREstimator}
import scala.collection.mutable
import soal.util.CollectionsUtil

/*
 * Created by plofgren on 5/12/15.
 */

class CrossPartitioningExperiments(graph: DirectedGraph[Node], teleportProb: Float,
                                    pageRanks: IndexedSeq[Double]) {
  val random = new Random()
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProb, random)
  val pprPrecomputor = new BidirectionalPPRPrecomputor(graph, teleportProb, random)

  def shardedRuntimeExperiment(pairCount: Int): Unit = {
    val shardCount = 100
    val minimumPPR = 1.0f / graph.nodeCount
    val maxResidual = 1.0e-4f // Chosen to minimize storage
    // math.sqrt(minimumPPR).toFloat // TODO
    val starts = Array.fill(pairCount) { random.nextInt(graph.maxNodeId + 1) }
    val targets = Array.fill(pairCount) {
      ExperimentUtils.sampleFromPageRank(graph, teleportProb).id
    }
    val (allEstimates, allResiduals) =  targets.par.map { t =>
      ExperimentUtils.printTime(s"precomputation for $t took ") {
        pprEstimator.computeContributions(t, maxResidual)
      }
    }.seq.unzip
    val allEstimatesMap = new mutable.HashMap[Int, collection.Map[Int, Float]]() ++
      (targets zip allEstimates)

    // allResiduals is a map t -> (map v -> r^t_v)
    // We want to shard it by v, to get shardCount maps, where
    // shardedResidualMaps(i) is a map t -> (map v -> r^t_v)
    //         where v % shardCount == i
    val shardedResidualMaps = Array.fill(shardCount) {
      new mutable.HashMap[Int, mutable.Map[Int, Float]]()
    }
    for ((t, rs) <- targets zip allResiduals) {
      for (m <- shardedResidualMaps) {
        m(t) = CollectionsUtil.efficientIntFloatMap()
      }
      for ((v, r) <- rs) {
        val shardIndex = v % shardCount
        shardedResidualMaps(shardIndex)(t)(v) = r
      }
    }

//    println(s"starts ${starts.toSeq}")
//    println(s"targets ${targets.toSeq}")
//    println(s"allResidual ${(targets zip allResiduals).mkString("\n  ")}")
//    println(s"shardedResidual ${((0 until shardCount) zip shardedResidualMaps).mkString("\n  ")}")


    val forwardProbabilities = starts.par.map { s =>
      ExperimentUtils.printTime(s"forward precomputation for $s took ") {
        pprPrecomputor.forwardPrecomputation(s, minimumPPR, maxResidual,
          relativeError = 0.1f) // TODO
      }
    }

    // TODO: This should be refactored to remove code duplication
    // forwardProbabilities is a map s -> (map v -> p_{s,t})
    // We want to shard it by v, to get shardCount maps, where
    // shardedResidualMaps(i) is a map t -> (map v -> r^t_v)
    //         where v % shardCount == i
    val shardedForwardProbabilities = Array.fill(shardCount) {
      new mutable.HashMap[Int, mutable.Map[Int, Float]]()
    }
    for ((s, pMap) <- starts zip forwardProbabilities) {
      for (m <- shardedForwardProbabilities) {
        m(s) = CollectionsUtil.efficientIntFloatMap()
      }
      for ((v, p) <- pMap) {
        val shardIndex = v % shardCount
        shardedForwardProbabilities(shardIndex)(s)(v) = p
      }
    }

//    println(s"forwardProbabilities ${(starts zip forwardProbabilities).mkString("\n  ")}")
//    println(s"shardedForward ${((0 until shardCount) zip shardedForwardProbabilities).mkString("\n  ")}")


    val (maxTimes, singleShardTimes, bidirectionalTimes) = ((starts zip targets).zipWithIndex
      map {
      case ((s, t), pairIndex) =>
        val (estimates, times) = ((0 until shardCount) map {
          shardIndex =>
            ExperimentUtils.time {
              CollectionsUtil.dotMaps(
                shardedForwardProbabilities(shardIndex)(s),
                shardedResidualMaps(shardIndex)(t)
              )
            }
        }).unzip
        val (estimate, summingTime) = ExperimentUtils.time {
          estimates.sum + allEstimatesMap(t)(s)
        }
        val (truePPR, bidirectionalTime)= ExperimentUtils.time {
          pprEstimator.estimatePPR(s, t, minimumPPR, 0.1f)
        }
        val (precomputationEstimate, singleShardTime) = ExperimentUtils.time {
          CollectionsUtil.dotMaps(
              forwardProbabilities(pairIndex),
              allResiduals(pairIndex)) +
            allEstimatesMap(t)(s)
        }
        // as sanity check, measure relative error
        val relativeError = math.abs(truePPR - estimate) / truePPR
        //println(s"s: $s   t: $t")
        //println(s"estimates: ${estimates.toSeq}")
        println(s"non-residual: ${allEstimatesMap(t)(s)} estimate: $estimate precompEstimate:" +
          s"$precomputationEstimate")
        println(s"truePPR: $truePPR   relativeError: $relativeError")
        println(s"meanTime: ${times.sum / shardCount} maxTime: ${times.max}")
        println(s"summingTime: $summingTime")
        (times.max, singleShardTime, bidirectionalTime)
    }).unzip3

    println("maxTimes = " + ExperimentUtils.toPythonList(maxTimes))
    println("singleShardTimes = " + ExperimentUtils.toPythonList(singleShardTimes))

    println("bidirectionalTimes = " + ExperimentUtils.toPythonList(bidirectionalTimes))
    val targetPageRanks = targets map pageRanks
    println("targetPageRanks = " + ExperimentUtils.toPythonList(targetPageRanks))
  }
}
