package soal.ppr.experiments

import java.io._
import java.lang.management.ManagementFactory

import com.twitter.cassovary.algorithms.linkanalysis.{PageRank, PageRankParams}
import com.twitter.cassovary.graph.StoredGraphDir.StoredGraphDir
import com.twitter.cassovary.graph.{DirectedGraph, MemoryMappedDirectedGraph, Node, StoredGraphDir}
import com.twitter.cassovary.util.NodeNumberer
import com.twitter.cassovary.util.io.AdjacencyListGraphReader

import scala.annotation.tailrec
import scala.collection.mutable.ArrayBuffer
import scala.util.Random

object ExperimentUtils {

  val nanosPerSecond = 1.0e9
  /**
   * Returns the result and running time of a piece of code, in seconds.
   */
  def oldTime[A](f: => A): (A, Double) = {
    val startTime = System.nanoTime
    val a = f
    val elapsedTime = (System.nanoTime - startTime) / nanosPerSecond
    (a, elapsedTime)
  }
  def time[A](f: => A): (A, Double) = {
    val startTime = ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime()
    val a = f
    val elapsedTime = (ManagementFactory.getThreadMXBean().getCurrentThreadCpuTime() - startTime) / nanosPerSecond
    (a, elapsedTime)
  }


  def printTime[A](message: String)(f: => A): A  = {
    val (result, runTime) = time {f}
    println(message + runTime)
    result
  }

  def readGraph(graphPath: String): DirectedGraph[Node] = {
    val graphExtension = graphPath.substring(math.max(0, graphPath.lastIndexOf('.')))
    if (graphExtension == ".dat") {
      new MemoryMappedDirectedGraph(new File(graphPath))
    } else {
      val filenameStart = graphPath.lastIndexOf('/') + 1
      val graphDirectory = graphPath.take(filenameStart)
      val graphFilename = graphPath.drop(filenameStart)
      println("loading graph:" + graphFilename)

      val reader = new AdjacencyListGraphReader(
        graphDirectory,
        graphFilename,
        new NodeNumberer.IntIdentity(),
        _.toInt) {
        override def storedGraphDir: StoredGraphDir = StoredGraphDir.BothInOut
      }
      val (graph, loadTime) = time {
        reader.toArrayBasedDirectedGraph(): DirectedGraph[Node]
      }

      println(s"loading graph of ${graph.nodeCount} nodes took $loadTime seconds.")
      val binaryGraphPath = graphPath.dropRight(4) + ".dat"
      println(s"converting graph to binary at $binaryGraphPath")
      ExperimentUtils.printTime("converting graph to binary took ") {
        MemoryMappedDirectedGraph.graphToFile(graph, new File(binaryGraphPath))
      }

      graph
    }
  }

  def using[B, A <: {def close(): Unit}](a: A)(f: A => B): B = {
    try {
      f(a)
    } finally {
      a.close()
    }
  }
  def loadOrComputePageRank(graph: DirectedGraph[Node], graphPath: String, teleportProb: Float): IndexedSeq[Double] = {
    val dotIndex = math.max(graphPath.lastIndexOf('.'), graphPath.lastIndexOf('/'))
    val pagerankFile = new File(graphPath.substring(0, dotIndex) + ".pagerank")
    if (pagerankFile.exists()) {
      using(new DataInputStream(new FileInputStream(pagerankFile))) {
        input =>
          val result = new Array[Double](graph.maxNodeId + 1)
          for (i <- 0 until (graph.maxNodeId + 1)) {
            result(i) = input.readDouble()
          }
          result
      }
    } else {
      val result = (new PageRank(graph, new PageRankParams(1.0 - teleportProb, maxIterations = Some(20)))).run().pageRank
      using(new DataOutputStream(new FileOutputStream(pagerankFile))) { output =>
        for(x <- result) {
          output.writeDouble(x)
        }
      }
      result
    }
  }

  def randomElement[A](xs: Seq[A]): A = {
    xs(Random.nextInt(xs.size))
  }

  /**
   * Assumes that nodeIds in graph are sequential.
   */
  /*def sampleFromPageRank(graph: DirectedGraph[Node], teleportProbability: Double): Node = {
    def randomNode(): Node =
      graph.getNodeById(Random.nextInt(graph.maxNodeId + 1)).get
    var currentNode = randomNode()
    while(Random.nextFloat() > teleportProbability) {
      currentNode = currentNode.randomOutboundNode match {
        case Some(uId) => graph.getNodeById(uId).get
        case None  => randomNode() // Teleport away from sink nodes
      }
    }
    currentNode
  }*/

  /* pass in Node sequence of graph because graph doesn't seem to have random access to nodes */
  def sampleFromPageRank(graph: DirectedGraph[Node], nodes: Seq[Node], teleportProbability: Double): Node
  = {
    var currentNode = randomElement(nodes)
    while(Random.nextFloat() > teleportProbability) {
      currentNode = currentNode.randomOutboundNode match {
        case Some(uId) => graph.getNodeById(uId).get
        case None  => randomElement(nodes) // Teleport away from sink nodes
      }
    }
    currentNode
  }

  @tailrec
  def randomNode(graph: DirectedGraph[Node]): Node = {
    val candidateNodeId = Random.nextInt(graph.maxNodeId)
    graph.getNodeById(candidateNodeId) match {
      case Some(node) => node
      case None => randomNode(graph)
    }
  }

  def sampleFromPageRank(graph: DirectedGraph[Node], teleportProbability: Double): Node
  = {
    var currentNode = randomNode(graph)
    while(Random.nextFloat() > teleportProbability) {
      currentNode = currentNode.randomOutboundNode match {
        case Some(uId) => graph.getNodeById(uId).get
        case None  => randomNode(graph) // Teleport away from sink nodes
      }
    }
    currentNode
  }

  def mean[A](xs: Seq[A])(implicit num: Numeric[A]): Double = num.toDouble(xs.sum) / xs.size
  //def sum[B >: A](implicit num: Numeric[B])

  def median[A](xs: Seq[A])(implicit ord: Ordering[A]) = {
    val sortedXs = xs.sorted
    sortedXs(xs.size / 2)  // Don't worry about case when xs.size is even
  }

  def toPythonList[A](m: Seq[A]): String = m.mkString("[", ", ", "]")

  def matrixToPythonList(m: Seq[Seq[Double]]): String =
      (m map toPythonList).mkString("[", ",\n", "]")

  def meanMatrix(ms: Seq[Seq[Seq[Double]]]): Seq[Seq[Double]] = {
    val n = ms.size // number of matrices
    val r = ms.head.size // number of rows
    val c = ms.head.head.size // number of columns
    val result = ArrayBuffer.fill(r)(ArrayBuffer.fill(c)(0.0))
    for (m <- ms) {
      for (i <- 0 until r) {
        for (j <- 0 until c) {
          result(i)(j) += m(i)(j) / n
        }
      }
    }
    result
  }

  def meanSeq(ms: Seq[Seq[Double]]): Seq[Double] = {
    val n = ms.size // number of seqs
    val r = ms.head.size // number of rows
    val result = ArrayBuffer.fill(r)(0.0)
    for (m <- ms) {
      for (i <- 0 until r) {
        result(i) += m(i) / n
      }
    }
    result
  }

  def medianSeq(ms: Seq[Seq[Double]]): Seq[Double] = {
    val c = ms.head.size // number of columns
    for (i <- 0 until c) yield {
      median(ms map (row => row(i)))
    }
  }
}
