package soal.ppr.experiments

import com.twitter.cassovary.graph.{Node, DirectedGraph}
import soal.markov.BidirectionalTransitionEstimator
import scala.util.Random
import scala.collection.mutable.ArrayBuffer

/*
 * Created by plofgren on 6/4/15.
 */

object HeatKernelAccuracyExperiments {
  def accuracyExperiment(graph: DirectedGraph[Node], trialCount: Int): (Double, Double, Double) = {
    val heatKernelT = 5
    val maxLength = (heatKernelT + 10 * math.sqrt(heatKernelT)).toInt // 10 stddev above mean
    val minimumProbability = 1.0 / graph.nodeCount
    val epsilons = List(minimumProbability / 2.0)
    val chernoffConstants = Seq(0.4)
    val mcWalkCount = (1.0 / minimumProbability).toInt

    def randomNode(): Node =  {
      graph.getNodeById(Random.nextInt(graph.maxNodeId)).getOrElse(randomNode())
    }
    val biRelativeErrors = new ArrayBuffer[Double]()
    val pushForwardRelativeErrors = new ArrayBuffer[Double]()
    val mcRelativeErrors = new ArrayBuffer[Double]()

    for ((epsilon, chernoffConstant) <- (epsilons zip chernoffConstants)) {
      val estimator = new BidirectionalTransitionEstimator(graph, heatKernelT, maxLength)
      val trueValues = new ArrayBuffer[Double]()
      val bidirectionalValues = new ArrayBuffer[Double]()
      val pushForwardValues = new ArrayBuffer[Double]()
      val monteCarloValues = new ArrayBuffer[Double]()

      (0 until trialCount) map { trialIndex =>
        val s = randomNode().id
        val monteCarloFromS = ExperimentUtils.printTime("computing MC took ") {
          estimator.monteCarloHeatKernel(s, mcWalkCount)
        }
        //def isSignificant(t: Int) = monteCarloFromS(t) > minimumProbability
        //val candidateTargets = (monteCarloFromS.keys filter isSignificant).toArray
        val pushForwardFromS = ExperimentUtils.printTime("computing pushForward value took ") {
          estimator.forwardPushHeatKernel(s, epsilon)
        }
        //val candidateTargets = (pushForwardFromS.keys).toArray
        val candidateTargets = (Seq(s) ++ monteCarloFromS.keys.filter {
          k => k >= 0 && monteCarloFromS(k) >= 1.99 * minimumProbability
        }).toArray

        println("# potential candidate targets: " + candidateTargets.size)
        var t = -1
        while (t == -1) {
          val candidate = candidateTargets(Random.nextInt(candidateTargets.size))
          val estimate = estimator.heatKernel(s, candidate, minimumProbability, 0.1,
            Some(0.1))
          if (estimate > minimumProbability) {
            t = candidate
          } else {
            println("rejecting candidate")
          }
        }
        val trueValue = ExperimentUtils.printTime("computing true value took ") {
          estimator.heatKernel(s, t, minimumProbability, 0.03, None)
        }
        val bidirectionalValue = ExperimentUtils.printTime("computing bidirectional value took ") {
          estimator.heatKernel(s, t, minimumProbability, 0.1, Some(chernoffConstant))
        }
        println(s"true\tbi\tforward\tmc: " +
          s"$trueValue\t$bidirectionalValue\t${pushForwardFromS(t)}\t${monteCarloFromS(t)}")
        trueValues += trueValue
        bidirectionalValues += bidirectionalValue
        pushForwardValues += pushForwardFromS(t)
        monteCarloValues += monteCarloFromS(t)

      }
      println(s"chernoffC=$chernoffConstant meanBiError: " +
        meanRelativeError(trueValues, bidirectionalValues))
      println("biErrors: " + relativeErrors(trueValues, bidirectionalValues))
      biRelativeErrors += meanRelativeError(trueValues, bidirectionalValues)

      println(s"epsilon=$epsilon meanForwardPushError: " +
        meanRelativeError(trueValues, pushForwardValues))
      println("forwardPushErrors: " + relativeErrors(trueValues, pushForwardValues))
      pushForwardRelativeErrors += meanRelativeError(trueValues, pushForwardValues)
      println(s"meanMCError: " +
        meanRelativeError(trueValues, monteCarloValues))
      mcRelativeErrors += meanRelativeError(trueValues, monteCarloValues)


      println("\n-----single-graph------")

      println("chernoffConstants: " + chernoffConstants)
      println("bidirectionalRelativeErrors: " + biRelativeErrors)
      println("epsilons = " + epsilons)
      println("pushForwardRelativeErrors: " + pushForwardRelativeErrors)
      println("mcdRelativeErrors: " + mcRelativeErrors)

    }
    (biRelativeErrors.head, pushForwardRelativeErrors.head, mcRelativeErrors.head)
  }

  def measureAccuracyOnGraphs(graphNames: Seq[String], trialCount: Int): Unit = {
    val accuracyTriples = new ArrayBuffer[(Double, Double, Double)]()
    for (graphName <- graphNames) yield {
      println("graph " + graphName)
      val graph = ExperimentUtils.readGraph(graphName)
      accuracyTriples += accuracyExperiment(graph, trialCount)
      val (times1, times2, times3) = accuracyTriples.unzip3
      println("------accuracy summary-------")
      for (accuracies <- List(times1, times2, times3)) {
        println(accuracies.mkString("\t"))
      }
      for (accuracies <- List(times1, times2, times3)) {
        println(ExperimentUtils.toPythonList(accuracies) + ",")
      }
    }
  }

  def meanRelativeError(trueValues: Seq[Double], estimatedValues: Seq[Double]): Double = {
    relativeErrors(trueValues, estimatedValues).sum / trueValues.size
  }

  def relativeErrors(trueValues: Seq[Double], estimatedValues: Seq[Double]): Seq[Double] = {
    (trueValues zip estimatedValues) map { case (t, e) =>
      math.abs(t - e) / t
    }
  }
}
