package soal.ppr.experiments

import com.twitter.cassovary.graph.{Node, DirectedGraph}
import scala.util.Random
import scala.collection.mutable.ArrayBuffer
import soal.markov.BidirectionalTransitionEstimator

object HeatKernelRuntimeExperiments {

  def measureRuntime(graph: DirectedGraph[Node], trialCount: Int): (Double, Double, Double) = {
    val heatKernelT = 5
    val maxLength = (heatKernelT + 10 * math.sqrt(heatKernelT)).toInt // 10 stddev above mean
    val minimumProbability = 1.0 / graph.nodeCount
    val epsilon = minimumProbability / 2.0
    val chernoffConstant = 0.4
    val mcWalkCount = (1.0 / minimumProbability).toInt

    def randomNode(): Node =
      graph.getNodeById(Random.nextInt(graph.maxNodeId)).getOrElse(randomNode())

    val starts = ArrayBuffer.fill(trialCount) {randomNode().id}
    val targets = ArrayBuffer.fill(trialCount) {randomNode().id}

    val biTimes = (starts zip targets).par.map { case (s, t) =>
      val estimator = new BidirectionalTransitionEstimator(graph, heatKernelT, maxLength)
      val (estimate, runTime) = ExperimentUtils.time {
        estimator.heatKernel(s, t, minimumProbability, 0.1, Some(chernoffConstant))
      }
      if (estimate == -123.0) println("don't optimize away")
      println("bi runTime: " + runTime)
      runTime
    }.seq

    val pushForwardTimes = starts.par.map { s =>
      val estimator = new BidirectionalTransitionEstimator(graph, heatKernelT, maxLength)
      val (estimate, runTime) = ExperimentUtils.time {
        estimator.forwardPushHeatKernel(s, epsilon)
      }
      if (estimate == -123.0) println("don't optimize away")
      println("pushForward runTime: " + runTime)
      runTime
    }.seq

    val mcTimes = starts.par.map { s =>
      val estimator = new BidirectionalTransitionEstimator(graph, heatKernelT, maxLength)
      val (estimate, runTime) = ExperimentUtils.time {
        estimator.monteCarloHeatKernel(s, mcWalkCount)
      }
      if (estimate == -123.0) println("don't optimize away")
      println("mcTimes runTime: " + runTime)
      runTime
    }.seq

    println("biTimes = " + ExperimentUtils.toPythonList(biTimes))
    println("pushForwardTimes = " + ExperimentUtils.toPythonList(pushForwardTimes))
    println("mcTimes = " + ExperimentUtils.toPythonList(mcTimes))

    val meanBiTime = ExperimentUtils.mean(biTimes)
    val meanPushForwardTime = ExperimentUtils.mean(pushForwardTimes)
    val meanMcTime = ExperimentUtils.mean(mcTimes)
    (meanBiTime, meanPushForwardTime, meanMcTime)
  }

  def measureRuntimeOnGraphs(graphNames: Seq[String], trialCount: Int): Unit = {
    val runtimeTriples = new ArrayBuffer[(Double, Double, Double)]()
      for (graphName <- graphNames) yield {
      println("graph " + graphName)
      val graph = ExperimentUtils.readGraph(graphName)
      runtimeTriples += measureRuntime(graph, trialCount)
      val (times1, times2, times3) = runtimeTriples.unzip3
      println("------summary-------")
      for (runtimes <- List(times1, times2, times3)) {
        println(runtimes.mkString("\t"))
      }
      for (runtimes <- List(times1, times2, times3)) {
        println(ExperimentUtils.toPythonList(runtimes) + ",")
      }
    }
  }
  /*
  def measureForwardReverseTime(startId: Int,
                                targetId: Int,
                                graph: DirectedGraph,
                                teleportProb: Double): (Float, Float) = {
    // This is copied and pasted from FastPPR. TODO: find a way of measure forward+backward time without copying+pasting
    val minimumPPR = 4.0f / graph.nodeCount
    val chernoffMultiplier = 7.0f
    // TODO: Add balance constant and tune in experiments, or implement balanced version
    val maxResidual = math.sqrt(minimumPPR / chernoffMultiplier).toFloat



    val ((estimates, residuals), reverseTime) = ExperimentUtils.time {
      pprEstimator.computeContributions(targetId, maxResidual)
    }
    val reverseSize = estimates.size + residuals.size
    println(s"reverseSize = $reverseSize reverseTime = $reverseTime")

    val walkCount = (chernoffMultiplier * maxResidual / minimumPPR).toInt
    val (estimate, forwardTime) = ExperimentUtils.time {
      var estimate = estimates(startId)
      for (walkIndex <- 0 until walkCount) {
        val v = pprEstimator.samplePPR(startId)
        estimate += residuals(v) / walkCount
      }
      estimate
    }


    println(s"walkCount = $walkCount  forwardTime = $forwardTime" )
    println("forwardTime = " + forwardTime)

    if (estimate == 1.2345f)
      println("Make sure the compiler doesn't optimize pprEstimate away.")
    (forwardTime.toFloat, reverseTime.toFloat)
  }
  */
}
