package soal.ppr.experiments

import com.twitter.cassovary.graph.node.DynamicNode
import com.twitter.cassovary.graph.{DynamicDirectedGraph, ArrayBasedDynamicDirectedGraph, Node, DirectedGraph}
import com.twitter.cassovary.graph.StoredGraphDir._
import scala.collection.mutable
import scala.util.Random
import soal.ppr.BidirectionalPPREstimator
import soal.util.{DiscreteDistribution, ConstantDistribution, CollectionsUtil, UniformDistribution}
import scala.collection.mutable.ArrayBuffer

/*
 * Created by plofgren on 6/16/15.
 */

class ShardedPageRankExperiments(graph: DirectedGraph[Node], teleportProb: Float) {
  val random = new Random
  val pprEstimator = new BidirectionalPPREstimator(graph, teleportProb)

  def run(pairCount: Int, trialCount: Int) {
    val minDegrees = 0 to 5
    println(s"pairCount = $pairCount")
    println(s"trialCount = $trialCount")
    val meanErrors = for (minDegree <- minDegrees) yield {
      println(s"starting iteration with minDegree $minDegree")
      val errors = for (trialI <- 0 until trialCount) yield
        shardedPPRAccuracy(minDegree, pairCount)
      println(s"minDegree $minDegree")
      println("errors = " + ExperimentUtils.toPythonList(errors))
      ExperimentUtils.mean(errors)
    }

    println("minDegrees " + ExperimentUtils.toPythonList(minDegrees))
    println("meanErrors = " + ExperimentUtils.toPythonList(meanErrors))
    //println("meanError = " + ExperimentUtils.mean(meanErrors))
  }

  def shardedPPRAccuracy(minDegree: Int, pairCount: Int): Float = {
    val n = graph.nodeCount
    val shardCount = 10
    val minPPR = 1.0f / n
    val uniformNodeDistribution = new DiscreteDistribution {
      override def sample(): Int = {
        val candidateId = random.nextInt(n)
        if (graph.getNodeById(candidateId).nonEmpty)
          candidateId
        else
          sample()
      }
      override def valuesAndProbabilities(): Iterable[(Int, Float)] = ???
    }
    val shards = ExperimentUtils.printTime("sharding graph took ") {
      shardGraph(graph, shardCount, minDegree)
    }
    val relativeErrors = new ArrayBuffer[Float]()
    while (relativeErrors.size < pairCount) {
      val start = random.nextInt(n)
      val target = pprEstimator.samplePPR(uniformNodeDistribution)
      val startNode = graph.getNodeById(start).get
      val targetNode = graph.getNodeById(target).get
      val trueValue = ExperimentUtils.printTime("computing true value took ") {
        pprEstimator.estimatePPR(start, target, minPPR, 0.01f)
      }
      if (trueValue > minPPR) {
        val shardEstimates = (shards.par map {shard =>
          // Add out-neighbors of start and in-neighbors of target
          val shardStartNode = shard.getOrCreateNode(start)
          val newStartOutNeighbors = startNode.outboundNodes() filterNot (shardStartNode.outboundNodes().contains)
          println(s"startNode has ${startNode.outboundCount}, of which ${shardStartNode.outboundCount} are in shard, and " +
            s"${newStartOutNeighbors.size} will be added now")
          shardStartNode.addOutBoundNodes(newStartOutNeighbors)
          val shardTargetNode = shard.getOrCreateNode(target)
          val newTargetInNeighbors = targetNode.inboundNodes() filterNot (shardTargetNode.inboundNodes().contains)
          println(s"targetNode has ${targetNode.inboundCount}, of which ${shardTargetNode.inboundCount} are in shard, and " +
            s"${newTargetInNeighbors.size} will be added now")
          shardTargetNode.addInBoundNodes(newTargetInNeighbors)
          //val shardUnionEndpoints = new GraphUnionNodes(shard, graph, Set(start, target))

          println("estimating ppr...")
          val shardEstimator = new BidirectionalPPREstimator(shard, teleportProb)
          val targetInboundSet = new mutable.HashSet[Int]() ++= targetNode.inboundNodes()
          def edgeWeight(u: Node, v: Node): Float =
            if (v.id == target)
              // The edge (u, v) has weight 1.0/shardCount.  The rest have equal weight 1.0
              (1.0f / shardCount) / (u.outboundCount - 1 + 1.0f / shardCount)
            else
              1.0f / u.outboundCount
          val shardEstimate = shardEstimator.estimatePPR(new ConstantDistribution(start), target,
            minPPR, 0.01f, Some(edgeWeight))

          newStartOutNeighbors map (shardStartNode.removeOutBoundNode(_))
          newTargetInNeighbors map (shardTargetNode.removeInBoundNode(_))
          shardEstimate
        }).seq
        val estimate = ExperimentUtils.mean(shardEstimates).toFloat
        val relativeError = math.abs(trueValue - estimate) / trueValue
        println("shard estimates: " + shardEstimates)
        println(s"true estimate relativeError: $trueValue\t$estimate\t$relativeError")
        relativeErrors += relativeError
      } else {
        println(s"rejecting ($start, $target) because ppr $trueValue is too small")
      }
    }
    println("relativeErrors = " + ExperimentUtils.toPythonList(relativeErrors))
    ExperimentUtils.mean(relativeErrors).toFloat
  }

  def shardGraph(graph: DirectedGraph[Node], shardCount: Int, minDegree: Int
      ): IndexedSeq[DynamicDirectedGraph[DynamicNode]] = {
    val shards = IndexedSeq.fill(shardCount)(new ArrayBasedDynamicDirectedGraph(BothInOut))
    for (u <- graph) {
      val permutedNeighbors = random.shuffle(u.outboundNodes())
      val d = u.outboundCount
      val subgraphOutDegree =  math.max(math.min(minDegree, d), (d + shardCount-1) / shardCount)
      for ((shard, shardI) <- shards.zipWithIndex) {
        for (i <- 0 until subgraphOutDegree) {
          shard.addEdge(u.id, permutedNeighbors((i + shardI * subgraphOutDegree)  % d))
        }
      }
    }
    shards
  }

  def shardGraphImplicitly(graph: DirectedGraph[Node], shardCount: Int, minDegree: Int
                  ): IndexedSeq[DynamicDirectedGraph[DynamicNode]] = {
    val shards = IndexedSeq.fill(shardCount)(new ArrayBasedDynamicDirectedGraph(BothInOut))
    for (u <- graph) {
      val permutedNeighbors = random.shuffle(u.outboundNodes())
      val d = u.outboundCount
      val subgraphOutDegree =  math.max(math.min(minDegree, d), (d + shardCount-1) / shardCount)
      for ((shard, shardI) <- shards.zipWithIndex) {
        for (i <- 0 until subgraphOutDegree) {
          shard.addEdge(u.id, permutedNeighbors((i + shardI * subgraphOutDegree)  % d))
        }
      }
    }
    shards
  }

}

/*
class GraphUnionNodes(subgraph: DirectedGraph[Node],
                   fullGraph: DirectedGraph[Node],
                   fullGraphNodeIds: collection.Set[Int]) extends DirectedGraph[Node] {
  override def getNodeById(id: Int): Option[Node] =
    if (fullGraphNodeIds.contains(id))
      fullGraph.getNodeById(id)
    else
      subgraph.getNodeById(id)

  override def iterator: Iterator[Node] = ???

  override def edgeCount: Long = ???

  override val storedGraphDir: StoredGraphDir = subgraph.storedGraphDir

  override def nodeCount: Int = ???
}
*/
